@@ -2,12 +2,15 @@ import sys
 import traceback
 import textwrap
 import warnings
-import types
-from contextlib import contextmanager
 
 import attr
 
-__all__ = ["MultiError"]
+# python traceback.TracebackException < 3.6.4 does not support unhashable exceptions
+# see https://github.com/python/cpython/pull/4014 for details
+if sys.version_info < (3, 6, 4):
+    exc_key = lambda exc: exc
+else:
+    exc_key = id
 
 ################################################################
 # MultiError
@@ -88,7 +91,7 @@ def _filter_impl(handler, root_exc):
             elif changed:
                 return MultiError(new_exceptions)
             else:
-                preserved.add(exc)
+                preserved.add(id(exc))
                 return exc
         else:
             new_exc = handler(exc)
@@ -98,7 +101,7 @@ def _filter_impl(handler, root_exc):
             return new_exc
 
     def push_tb_down(tb, exc, preserved):
-        if exc in preserved:
+        if id(exc) in preserved:
             return
         new_tb = concat_tb(tb, exc.__traceback__)
         if isinstance(exc, MultiError):
@@ -171,26 +174,36 @@ class MultiError(BaseException):
 
     """
 
+    def __init__(self, exceptions):
+        # Avoid recursion when exceptions[0] returned by __new__() happens
+        # to be a MultiError and subsequently __init__() is called.
+        if hasattr(self, "exceptions"):
+            # __init__ was already called on this object
+            assert len(exceptions) == 1 and exceptions[0] is self
+            return
+        self.exceptions = exceptions
+
     def __new__(cls, exceptions):
         exceptions = list(exceptions)
         for exc in exceptions:
             if not isinstance(exc, BaseException):
-                raise TypeError(
-                    "Expected an exception object, not {!r}".format(exc)
-                )
+                raise TypeError("Expected an exception object, not {!r}".format(exc))
         if len(exceptions) == 1:
+            # If this lone object happens to itself be a MultiError, then
+            # Python will implicitly call our __init__ on it again.  See
+            # special handling in __init__.
             return exceptions[0]
         else:
-            self = BaseException.__new__(cls)
-            self.exceptions = exceptions
-            return self
+            # The base class __new__() implicitly invokes our __init__, which
+            # is what we want.
+            #
+            # In an earlier version of the code, we didn't define __init__ and
+            # simply set the `exceptions` attribute directly on the new object.
+            # However, linters expect attributes to be initialized in __init__.
+            return BaseException.__new__(cls, exceptions)
 
     def __str__(self):
-        def format_child(exc):
-            #return "{}: {}".format(exc.__class__.__name__, exc)
-            return repr(exc)
-
-        return ", ".join(format_child(exc) for exc in self.exceptions)
+        return ", ".join(repr(exc) for exc in self.exceptions)
 
     def __repr__(self):
         return "<MultiError: {}>".format(self)
@@ -266,16 +279,20 @@ if have_tproxy:
             # no missing test we could add, and no value in coverage nagging
             # us about adding one.
             if operation.opname in [
-                "__getattribute__", "__getattr__"
+                "__getattribute__",
+                "__getattr__",
             ]:  # pragma: no cover
                 if operation.args[0] == "tb_next":
                     return tb_next
             return operation.delegate()
 
         return tputil.make_proxy(controller, type(base_tb), base_tb)
+
+
 else:
     # ctypes it is
     import ctypes
+
     # How to handle refcounting? I don't want to use ctypes.py_object because
     # I don't understand or trust it, and I don't want to use
     # ctypes.pythonapi.Py_{Inc,Dec}Ref because we might clash with user code
@@ -360,7 +377,7 @@ def traceback_exception_init(
     limit=None,
     lookup_lines=True,
     capture_locals=False,
-    _seen=None
+    _seen=None,
 ):
     if _seen is None:
         _seen = set()
@@ -374,14 +391,14 @@ def traceback_exception_init(
         limit=limit,
         lookup_lines=lookup_lines,
         capture_locals=capture_locals,
-        _seen=_seen
+        _seen=_seen,
     )
 
     # Capture each of the exceptions in the MultiError along with each of their causes and contexts
     if isinstance(exc_value, MultiError):
         embedded = []
         for exc in exc_value.exceptions:
-            if exc not in _seen:
+            if exc_key(exc) not in _seen:
                 embedded.append(
                     traceback.TracebackException.from_exception(
                         exc,
@@ -390,7 +407,7 @@ def traceback_exception_init(
                         capture_locals=capture_locals,
                         # copy the set of _seen exceptions so that duplicates
                         # shared between sub-exceptions are not omitted
-                        _seen=set(_seen)
+                        _seen=set(_seen),
                     )
                 )
         self.embedded = embedded
@@ -398,7 +415,7 @@ def traceback_exception_init(
         self.embedded = []
 
 
-traceback.TracebackException.__init__ = traceback_exception_init
+traceback.TracebackException.__init__ = traceback_exception_init  # type: ignore
 traceback_exception_original_format = traceback.TracebackException.format
 
 
@@ -407,12 +424,10 @@ def traceback_exception_format(self, *, chain=True):
 
     for i, exc in enumerate(self.embedded):
         yield "\nDetails of embedded exception {}:\n\n".format(i + 1)
-        yield from (
-            textwrap.indent(line, " " * 2) for line in exc.format(chain=chain)
-        )
+        yield from (textwrap.indent(line, " " * 2) for line in exc.format(chain=chain))
 
 
-traceback.TracebackException.format = traceback_exception_format
+traceback.TracebackException.format = traceback_exception_format  # type: ignore
 
 
 def trio_excepthook(etype, value, tb):
@@ -420,21 +435,22 @@ def trio_excepthook(etype, value, tb):
         sys.stderr.write(chunk)
 
 
-IPython_handler_installed = False
-warning_given = False
+monkeypatched_or_warned = False
+
 if "IPython" in sys.modules:
     import IPython
+
     ip = IPython.get_ipython()
     if ip is not None:
         if ip.custom_exceptions != ():
             warnings.warn(
                 "IPython detected, but you already have a custom exception "
-                "handler installed. I'll skip installing trio's custom "
+                "handler installed. I'll skip installing Trio's custom "
                 "handler, but this means MultiErrors will not show full "
                 "tracebacks.",
-                category=RuntimeWarning
+                category=RuntimeWarning,
             )
-            warning_given = True
+            monkeypatched_or_warned = True
         else:
 
             def trio_show_traceback(self, etype, value, tb, tb_offset=None):
@@ -443,15 +459,45 @@ if "IPython" in sys.modules:
                 trio_excepthook(etype, value, tb)
 
             ip.set_custom_exc((MultiError,), trio_show_traceback)
-            IPython_handler_installed = True
+            monkeypatched_or_warned = True
 
 if sys.excepthook is sys.__excepthook__:
     sys.excepthook = trio_excepthook
-else:
-    if not IPython_handler_installed and not warning_given:
-        warnings.warn(
-            "You seem to already have a custom sys.excepthook handler "
-            "installed. I'll skip installing trio's custom handler, but this "
-            "means MultiErrors will not show full tracebacks.",
-            category=RuntimeWarning
-        )
+    monkeypatched_or_warned = True
+
+# Ubuntu's system Python has a sitecustomize.py file that import
+# apport_python_hook and replaces sys.excepthook.
+#
+# The custom hook captures the error for crash reporting, and then calls
+# sys.__excepthook__ to actually print the error.
+#
+# We don't mind it capturing the error for crash reporting, but we want to
+# take over printing the error. So we monkeypatch the apport_python_hook
+# module so that instead of calling sys.__excepthook__, it calls our custom
+# hook.
+#
+# More details: https://github.com/python-trio/trio/issues/1065
+if getattr(sys.excepthook, "__name__", None) == "apport_excepthook":
+    import apport_python_hook
+
+    assert sys.excepthook is apport_python_hook.apport_excepthook
+
+    # Give it a descriptive name as a hint for anyone who's stuck trying to
+    # debug this mess later.
+    class TrioFakeSysModuleForApport:
+        pass
+
+    fake_sys = TrioFakeSysModuleForApport()
+    fake_sys.__dict__.update(sys.__dict__)
+    fake_sys.__excepthook__ = trio_excepthook  # type: ignore
+    apport_python_hook.sys = fake_sys
+
+    monkeypatched_or_warned = True
+
+if not monkeypatched_or_warned:
+    warnings.warn(
+        "You seem to already have a custom sys.excepthook handler "
+        "installed. I'll skip installing Trio's custom handler, but this "
+        "means MultiErrors will not show full tracebacks.",
+        category=RuntimeWarning,
+    )
